/* Copyright © 2012, 2013 Matthew Champion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
 * Neither the name of mattunderscore.com nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.filter.contentnegotiation;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.RequestDispatcher;

import com.mattunderscore.http.headers.QContentType;
import com.mattunderscore.http.headers.QLanguage;
import com.mattunderscore.http.headers.UnParsableHeaderException;

import com.mattunderscore.filter.GenericHTTPFilter;

import com.mattunderscore.filter.contentnegotiation.parser.CrossProduct;
import com.mattunderscore.filter.contentnegotiation.parser.Header;
import com.mattunderscore.filter.contentnegotiation.parser.HeaderPart;
import com.mattunderscore.filter.contentnegotiation.parser.Variant;
import com.mattunderscore.filter.contentnegotiation.parser.Redirect;
import com.mattunderscore.filter.contentnegotiation.parser.Output;
import com.mattunderscore.filter.contentnegotiation.variantsource.CachedVariantSource;
import com.mattunderscore.filter.contentnegotiation.variantsource.ConfigurableVariantSource;
import com.mattunderscore.filter.contentnegotiation.variantsource.VariantSource;
import com.mattunderscore.filter.contentnegotiation.variantsource.VariantSourceForFilter;

/**
 * A {@link javax.servlet.Filter} that can be used for content negotiation of HTTP requests. Loads the
 * filter configuration from the file negotiation.xml. Supports the accept
 * header and the language header.
 * 
 * @author Matt Champion
 * @since 0.0.13
 */
public class ContentNegotiationFilter extends GenericHTTPFilter
{
    /**
     * The parameter name of the variant source class parameter.
     */
    private static final String VARIANT_SOURCE_PARAMETER = "variantSource";
    
    /**
     * Logger used of information or warnings
     */
    private static Logger log = Logger
            .getLogger("com.mattunderscore.filter.contentnegotiation");
    /**
     * List of variants constructed from the negotiation.xml
     */
    private Map<String,String> parameterMap = new HashMap<String,String>();
    private String variantSourceClass = "com.mattunderscore.filter.contentnegotiation.variantsource.FilterVariantXMLSource";
    private VariantSource source;

    @Override
    public void configureFilter()
    {
        try
        {
            @SuppressWarnings("unchecked")
            Enumeration<String> parameterNames = getFilterConfig()
                    .getInitParameterNames();
            while (parameterNames.hasMoreElements())
            {
                String parameterName = parameterNames.nextElement();
                if (VARIANT_SOURCE_PARAMETER.equals(parameterName))
                {
                    variantSourceClass = getFilterConfig().getInitParameter(
                            parameterName);
                }
                parameterMap.put(parameterName, getFilterConfig().getInitParameter(parameterName));
                log.info("Config: " + parameterName + ": " + getFilterConfig().getInitParameter(parameterName));
            }
            Class<?> sourceClass = Class.forName(variantSourceClass);
            Object sourceObject = sourceClass.newInstance();
            if (sourceObject instanceof VariantSourceForFilter)
            {
                VariantSourceForFilter confSource = (VariantSourceForFilter) sourceObject;
                confSource.configureSource(getFilterConfig());
                source = new CachedVariantSource(confSource);
            }
            else if (sourceObject instanceof ConfigurableVariantSource)
            {
                ConfigurableVariantSource confSource = (ConfigurableVariantSource) sourceObject;
                confSource.configureSource(parameterMap);
                source = new CachedVariantSource(confSource);
            }
            else if (sourceObject instanceof VariantSource)
            {
                source = new CachedVariantSource((VariantSource)sourceObject);
            }
            else
            {
                log.warning("No source configured.");
            }
        }
        catch (Throwable t)
        {
            log.log(Level.WARNING,"ContentNegotiationFilter configuration error",t);
        }
    }

    @Override
    public void doHTTPFilter(HttpServletRequest request,
            HttpServletResponse response, FilterChain chain)
            throws IOException, ServletException
    {
        try
        {
            List<Variant> variants = source.getVariants();
            if (variants != null)
            {
                List<Variant> matchedVariants = getUrlVariants(variants,
                        request.getServletPath());
                if (matchedVariants.size() > 0)
                {
                    try
                    {
                        Variant bestVariant = getBestVariant(request,
                                matchedVariants);
                        handleOutput(request, response, bestVariant);
                        return;
                    }
                    catch (UnParsableHeaderException e)
                    {
                        log.log(Level.WARNING, e.getMessage(), e);
                    }
                    catch (NoMatchingRuleException e)
                    {
                        log.warning(e.getMessage());
                    }
                }
                else
                {
                    log.warning("No Variants on path:"
                            + request.getServletPath());
                }
            }
            chain.doFilter(request, response);
        }
        catch (Throwable t)
        {
            StringBuilder sb = new StringBuilder(100);
            sb.append("Request failed:\n");
            for (String key : parameterMap.keySet())
            {
                sb.append("conf: ");
                sb.append(key);
                sb.append(", ");
                sb.append(parameterMap.get(key));
                sb.append("\n");
            }
            log.warning(sb.toString());
            throw new ServletException(t);
        }
    }

    /**
     * Get the variants that match the URL pattern.
     * 
     * @param variants
     *            The list of all Variants
     * @param path
     *            The HTTP request path
     * @return The list of Variants that the request might match
     * @since 0.0.13
     */
    private List<Variant> getUrlVariants(final List<Variant> variants,
            final String path)
    {
        List<Variant> matchedVariants = new ArrayList<Variant>();
        for (Variant variant : variants)
        {
            if (variant.requestPathMatchesVariant(path))
            {
                matchedVariants.add(variant);
            }
        }
        return matchedVariants;
    }

    /**
     * Create the cross products that will be used to find the best content
     * type.
     * 
     * @param contentTypes
     *            The content types of the accept header of the HTTP request
     * @param variants
     *            The Variants that might apply to the request
     * @return The list of CrossProducts for the variants and the content types
     * @throws UnParsableHeaderException
     * @since 0.0.13
     */
    private List<CrossProduct> createContentTypeCrossProducts(
            List<QContentType> contentTypes, List<Variant> variants)
            throws UnParsableHeaderException
    {
        List<CrossProduct> products = new ArrayList<CrossProduct>();
        for (QContentType contentType : contentTypes)
        {
            for (Variant variant : variants)
            {
                for (QContentType variantType : variant.getQContentTypes())
                {
                    if (contentType.sameType(variantType))
                    {
                        products.add(new CrossProduct(contentType, variantType,
                                variant));
                    }
                }
            }
        }
        return products;
    }

    /**
     * Create the cross products that will be used to find the best language.
     * 
     * @param langs
     *            The languages of the accept-language header of the HTTP
     *            request
     * @param variants
     *            The Variants that might apply to the request
     * @return The list of cross products between the variants and the
     *         languages.
     * @throws UnParsableHeaderException
     * @since 0.0.13
     */
    private List<CrossProduct> createLanguageCrossProducts(
            List<QLanguage> langs, List<Variant> variants)
            throws UnParsableHeaderException
    {
        List<CrossProduct> products = new ArrayList<CrossProduct>();
        for (QLanguage lang : langs)
        {
            for (Variant variant : variants)
            {
                for (QLanguage variantLang : variant.getQLanguages())
                {
                    if (lang.sameLangauge(variantLang))
                    {
                        products.add(new CrossProduct(lang, variantLang,
                                variant));
                    }
                }
            }
        }
        return products;
    }

    /**
     * Returns the variants that have the best cross product.
     * 
     * @param cps
     *            List of CrossProducts
     * @return List of Variants
     * @since 0.0.13
     */
    private List<Variant> getBestVariants(List<CrossProduct> cps)
    {
        List<Variant> bvs = new ArrayList<Variant>();
        double qualifier = 0.0;
        for (CrossProduct cp : cps)
        {
            if (cp.getQualifier() > qualifier)
            {
                bvs = new ArrayList<Variant>();
                bvs.add(cp.getVariant());
                qualifier = cp.getQualifier();
            }
            else if (cp.getQualifier() == qualifier && qualifier > 0.0)
            {
                bvs.add(cp.getVariant());
            }
        }
        return bvs;
    }

    /**
     * Return the best variant of a list of variants.
     * 
     * @param request
     *            The servlet request, parsed to get requirements of request.
     * @param variants
     *            The list of Variants that might be used to resolve the
     *            request.
     * @return The best variant for the request.
     * @throws NoMatchingRuleException
     * @throws UnParsableHeaderException
     * @throws IOException
     * @since 0.0.13
     */
    private Variant getBestVariant(HttpServletRequest request,
            List<Variant> variants) throws NoMatchingRuleException,
            UnParsableHeaderException, IOException
    {
        List<QContentType> contentTypes = QContentType
                .getRequestContentTypes(request);
        List<CrossProduct> contentTypesCrossProduct = createContentTypeCrossProducts(
                contentTypes, variants);
        List<Variant> bestContentTypeVariants = getBestVariants(contentTypesCrossProduct);
        if (bestContentTypeVariants.size() == 0)
        {
            throw new NoMatchingRuleException();
        }
        else if (bestContentTypeVariants.size() == 1)
        {
            // Return the only acceptable variant
            return bestContentTypeVariants.get(0);
        }
        List<QLanguage> languages = QLanguage.getRequestLanguages(request);
        List<CrossProduct> languagesCrossProducts = createLanguageCrossProducts(
                languages, bestContentTypeVariants);
        List<Variant> bestLanguageVariants = getBestVariants(languagesCrossProducts);
        if (bestLanguageVariants.size() == 0)
        {
            // Return an acceptable content type variant
            return bestContentTypeVariants.get(0);
        }
        else if (bestLanguageVariants.size() == 1)
        {
            // Return the only acceptable variant
            return bestLanguageVariants.get(0);
        }
        // Return first of the remaining variants
        return bestLanguageVariants.get(0);
    }

    /**
     * Perform substitution of submatches for wildcards on redirect.
     * 
     * @param inPath
     *            The path of the request
     * @param stringPattern
     *            The URL pattern of the variant, with wild cards
     * @param outPath
     *            The destination path, with back references
     * @return The destination path with back references replaced with the wild
     *         card values
     * @throws BadSubstitution
     * @since 0.0.13
     */
    private String substitueMatch(String inPath, String stringPattern,
            String outPath) throws BadSubstitution
    {
        Pattern pathPattern = Pattern.compile(stringPattern);
        Matcher pathMatcher = pathPattern.matcher(inPath);
        if (pathMatcher.find())
        {
            if (pathMatcher.groupCount() < 1)
            {
                return outPath;
            }
            else
            {
                Pattern backRefPattern = Pattern.compile("\\$([0-9])");
                Matcher backRefMatcher = backRefPattern.matcher(outPath);
                String newString = outPath;
                while (backRefMatcher.find())
                {
                    String stringIndex = backRefMatcher.group(1);
                    int index = Integer.parseInt(stringIndex);
                    if (index > pathMatcher.groupCount())
                    {
                        throw new BadSubstitution();
                    }
                    String subValue = pathMatcher.group(index);
                    newString = newString.replace(backRefMatcher.group(),
                            subValue);
                }
                return newString;
            }
        }
        else
        {
            throw new BadSubstitution();
        }
    }

    /**
     * Perform the output behaviour of the variant. This forwards the request or
     * flushes the buffer.
     * 
     * @param request
     *            The servlet request, used for HTTP information
     * @param response
     *            The servlet response, used to write the response
     * @param variant
     *            The variant, used to determine the response
     * @throws IOException
     * @throws ServletException
     * @since 0.0.13
     */
    private void handleOutput(HttpServletRequest request,
            HttpServletResponse response, Variant variant) throws IOException,
            ServletException
    {
        boolean varyContentType = false;
        boolean varyLanguage = false;
        if (variant.getContentType() != null)
        {
            varyContentType = true;
        }
        if (variant.getLanguage() != null)
        {
            varyLanguage = true;
        }
        if (varyContentType && varyLanguage)
        {
            response.addHeader("Vary", "Accept,Accept-Language");
        }
        else if (varyContentType)
        {
            response.addHeader("Vary", "Accept");
        }
        else if (varyLanguage)
        {
            response.addHeader("Vary", "Accept-Language");
        }
        Output output = variant.getOutput();
        if (output.getForward() != null)
        {
            RequestDispatcher rd = request.getRequestDispatcher(output
                    .getForward());
            log.info("Forwading request to " + output.getForward());
            rd.forward(request, response);
            return;
        }
        else if (output.getHeader() != null)
        {
            Header header = output.getHeader();
            response.setStatus(header.getCode());
            for (HeaderPart part : header.getPart())
            {
                String value = request.getContextPath() + part.getValue();
                response.addHeader(part.getName(), value);
            }
            log.info("Returning custom header with code: " + header.getCode());
            response.flushBuffer();
            return;
        }
        else if (output.getRedirect() != null)
        {
            Redirect redirect = output.getRedirect();
            response.setStatus(redirect.getCode());
            String newLocation;
            try
            {
                newLocation = substitueMatch(request.getServletPath(),
                        variant.getUrlPattern(), request.getContextPath()
                                + redirect.getLocation());
            }
            catch (BadSubstitution ex)
            {
                newLocation = request.getContextPath() + redirect.getLocation();
            }
            log.info("Redirecting " + request.getServletPath() + " to "
                    + newLocation + " with " + redirect.getCode());
            response.addHeader("location", request.getContextPath()
                    + newLocation);
            response.flushBuffer();
            return;
        }
        else if (output.isFail() != null)
        {
            if (output.isFail())
            {
                log.info("Returning 406 header");
                response.setStatus(406);
                response.flushBuffer();
                return;
            }
        }
    }

    /**
     * When there is no Variant that can be matched to the request. This should
     * not be exposed outside of this class, handling of this exception should
     * be done in this class.
     * 
     * @author Matt Champion
     * @since 0.0.13
     */
    @SuppressWarnings("serial")
    private class NoMatchingRuleException extends Exception
    {
        public NoMatchingRuleException()
        {
            super("No matching Variant");
        }
    }

    /**
     * When there is a problem with the substitution of wild cards and back
     * references. This should not be exposed outside of this class, handling of
     * this exception should be done in this class.
     * 
     * @author Matt Champion
     * @since 0.0.13
     */
    @SuppressWarnings("serial")
    private class BadSubstitution extends Exception
    {
        public BadSubstitution()
        {
            super("The back reference cannot be substituted");
        }
    }
}
