/* Copyright © 2012 Matthew Champion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of mattunderscore.com nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.filter.contentnegotiation.parser;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.SchemaFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

/**
 * DOM based parser for negotiation XML. Replaces JAXB to eliminate GPL license
 * constraint. It is intended that this will be replaced with a SAX parser.
 * 
 * @author Matt Champion
 * @since 0.0.13
 * @deprecated Deprecated since 0.2.0, replaced by 
 */
public class NegotiationDOMParser implements NegotiationXMLParser
{

    private static final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    private DocumentBuilder docBuilder;

    static {
        SchemaFactory schemaFactory = SchemaFactory.newInstance("http://www.w3.org/2001/XMLSchema");
        factory.setValidating(false);
        factory.setNamespaceAware(true);
        InputStream is = NegotiationDOMParser.class.getResourceAsStream("/negotiation.xsd");
        StreamSource ss = new StreamSource(is);
        try
        {
            factory.setSchema(schemaFactory.newSchema(ss));
        }
        catch (SAXException e)
        {
            e.printStackTrace();
        }
    }
    
    public NegotiationDOMParser() throws ParserConfigurationException
    {
        docBuilder = factory.newDocumentBuilder();
        docBuilder.setErrorHandler(new BasicErrorHandeler());
    }

    /**
     * Parse an input stream to a list of variants.
     * @since 0.0.13
     */
    @Override
    public List<Variant> parse(final InputStream iStream)
            throws ParserException
    {
        try
        {
            Document doc = docBuilder.parse(iStream);
            return parse(doc);
        }
        catch (SAXException ex)
        {
            throw new ParserException("SAXException: " + ex);
        }
        catch (IOException ex)
        {
            throw new ParserException("IOException: " + ex);
        }
    }

    private List<Variant> parse(final Document doc) throws ParserException
    {
        doc.getDocumentElement().normalize();
        Element root = doc.getDocumentElement();
        String nodeName = root.getNodeName();
        if ("variants".equals(nodeName))
        {
            List<Variant> variants = new ArrayList<Variant>();
            NodeList children = root.getChildNodes();
            for (int i = 0; i < children.getLength(); i++)
            {
                Node n = children.item(i);
                if (n.getNodeType() == Node.ELEMENT_NODE)
                {
                    Element child = (Element) n;
                    variants.add(parseVariant(child));
                }
            }
            return variants;
        }
        else
        {
            throw new ParserException("parse: Not Variants");
        }
    }

    private Variant parseVariant(final Element variantElement)
            throws ParserException
    {
        Variant variant = new Variant();
        if ("variant".equals(variantElement.getNodeName()))
        {
            NodeList children = variantElement.getChildNodes();
            for (int i = 0; i < children.getLength(); i++)
            {
                Node n = children.item(i);
                if (n.getNodeType() == Node.ELEMENT_NODE)
                {
                    Element child = (Element) n;
                    if ("urlPattern".equals(child.getNodeName()))
                    {
                        variant.setUrlPattern(child.getTextContent());
                    }
                    else if ("content-type".equals(child.getNodeName()))
                    {
                        variant.setContentType(child.getTextContent());
                    }
                    else if ("language".equals(child.getNodeName()))
                    {
                        variant.setLanguage(child.getTextContent());
                    }
                    else if ("output".equals(child.getNodeName()))
                    {
                        variant.setOutput(parseOutput(child));
                    }
                    else
                    {
                        throw new ParserException(
                                "parseVariant: Not part of variant");
                    }
                }
            }
        }
        else
        {
            throw new ParserException("parseVariant: Not Variant");
        }
        return variant;
    }

    private Output parseOutput(final Element outputElement)
            throws ParserException
    {
        Output output = new Output();
        if ("output".equals(outputElement.getNodeName()))
        {
            NodeList children = outputElement.getChildNodes();
            for (int i = 0; i < children.getLength(); i++)
            {
                Node n = children.item(i);
                if (n.getNodeType() == Node.ELEMENT_NODE)
                {
                    Element child = (Element) n;
                    if ("fail".equals(child.getNodeName()))
                    {
                        output.setFail(true);
                    }
                    else if ("header".equals(child.getNodeName()))
                    {
                        output.setHeader(parseHeader(child));
                    }
                    else if ("redirect".equals(child.getNodeName()))
                    {
                        output.setRedirect(parseRedirect(child));
                    }
                    else if ("forward".equals(child.getNodeName()))
                    {
                        output.setForward(child.getTextContent());
                    }
                    else
                    {
                        throw new ParserException(
                                "parseOutput: Not part of output");
                    }
                }
            }
        }
        else
        {
            throw new ParserException("parseOutput: Not Output");
        }
        return output;
    }

    private Header parseHeader(final Element headerElement)
            throws ParserException
    {
        Header header = new Header();
        List<HeaderPart> list = new ArrayList<HeaderPart>();
        header.setPart(list);
        if ("header".equals(headerElement.getNodeName()))
        {
            NodeList children = headerElement.getChildNodes();
            for (int i = 0; i < children.getLength(); i++)
            {
                Node n = children.item(i);
                if (n.getNodeType() == Node.ELEMENT_NODE)
                {
                    Element child = (Element) n;
                    if ("part".equals(child.getNodeName()))
                    {
                        list.add(parseHeaderPart(child));
                    }
                    else
                    {
                        throw new ParserException(
                                "parseHeader: Not part of header");
                    }
                }
            }
        }
        else
        {
            throw new ParserException("parseHeader: Not Header");
        }
        return header;
    }

    private HeaderPart parseHeaderPart(final Element headerPartElement)
            throws ParserException
    {
        HeaderPart headerPart = new HeaderPart();
        if ("part".equals(headerPartElement.getNodeName()))
        {
            NodeList children = headerPartElement.getChildNodes();
            for (int i = 0; i < children.getLength(); i++)
            {
                Node n = children.item(i);
                if (n.getNodeType() == Node.ELEMENT_NODE)
                {
                    Element child = (Element) n;
                    if ("name".equals(child.getNodeName()))
                    {
                        headerPart.setName(child.getTextContent());
                    }
                    else if ("value".equals(child.getNodeName()))
                    {
                        headerPart.setValue(child.getTextContent());
                    }
                    else
                    {
                        throw new ParserException(
                                "parseHeaderPart: Not part of header part");
                    }
                }
            }
        }
        else
        {
            throw new ParserException("parseHeaderPart: not HeaderPart");
        }
        return headerPart;
    }

    private Redirect parseRedirect(final Element redirectElement)
            throws ParserException
    {
        Redirect redirect = new Redirect();
        if ("redirect".equals(redirectElement.getNodeName()))
        {
            NodeList children = redirectElement.getChildNodes();
            for (int i = 0; i < children.getLength(); i++)
            {
                Node n = children.item(i);
                if (n.getNodeType() == Node.ELEMENT_NODE)
                {
                    Element child = (Element) n;
                    if ("code".equals(child.getNodeName()))
                    {
                        try
                        {
                            redirect.setCode(Integer.parseInt(child
                                    .getTextContent()));
                        }
                        catch (NumberFormatException ex)
                        {
                            throw new ParserException(
                                    "parseRedirect: Not number");
                        }
                    }
                    else if ("location".equals(child.getNodeName()))
                    {
                        redirect.setLocation(child.getTextContent());
                    }
                    else
                    {
                        throw new ParserException(
                                "parseRedirect: Not part of redirect");
                    }
                }
            }
        }
        else
        {
            throw new ParserException("parseRedirect: Not Redirect");
        }
        return redirect;
    }
    
    private class BasicErrorHandeler implements ErrorHandler
    {
        public void error(SAXParseException ex) throws SAXException
        {
            throw ex;
        }

        public void fatalError(SAXParseException ex) throws SAXException
        {
            throw ex;
        }

        public void warning(SAXParseException ex) throws SAXException
        {
            throw ex;
        }
    }
}
