/* Copyright © 2013 Matthew Champion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
 * Neither the name of mattunderscore.com nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.filter.contentnegotiation.variantsource;

import java.util.Collections;
import java.util.List;

import com.mattunderscore.filter.contentnegotiation.parser.Variant;

/**
 * Supports caching variant sources by providing a wrapper around them. This will only call the
 * underlying variant source when {@link #isCacheEmpty()} returns true. The cache starts empty and can be
 * cleared. The variant list this returns is unmodifiable.
 * 
 * @author Matt Champion
 * @since 0.2.0
 */
public class CachedVariantSource implements VariantSource
{
    private final VariantSource source;
    private List<Variant> variants;

    /**
     * Constructs a cached version of the variant source.
     * 
     * @param source
     *            Underlying variant source
     * @since 0.2.0
     */
    public CachedVariantSource(VariantSource source)
    {
        this.source = source;
    }

    @Override
    public List<Variant> getVariants() throws VariantSourceException
    {
        if (isCacheEmpty())
        {
            variants = Collections.unmodifiableList(source.getVariants());
        }
        return variants;
    }

    /**
     * Returns true when there are no cached variants.
     * 
     * @return True when there are no cached variants.
     * @since 0.2.0
     */
    public boolean isCacheEmpty()
    {
        return variants == null;
    }

    /**
     * Clears the cache of any variants. The next call to {@link #getVariants()} will cause the
     * underlying source to provide a new list.
     * 
     * @since 0.2.0
     */
    public void clearCache()
    {
        variants = null;
    }

}
