/* Copyright © 2013 Matthew Champion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
 * Neither the name of mattunderscore.com nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MATTHEW CHAMPION BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

package com.mattunderscore.filter.contentnegotiation.variantsource;

import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;

import com.mattunderscore.filter.contentnegotiation.parser.Variant;

/**
 * A variant source that parses XML for variants and is configured from a filter. It is intended to
 * be called from filters and is configured with a filter config. Any configuration this needs
 * should be set in the filter init-params of the web.xml. The {@link VariantSAXHandler} that is
 * used to parse the XML as be configured.
 * 
 * @author Matt Champion
 * @since 0.2.0
 */
public class FilterVariantXMLSource implements VariantSourceForFilter
{
    /**
     * Name of the parameter that identifies the SAX hander to use.
     */
    public static final String SAX_HANDLER = "saxHandlerClass";
    public static final String XML_INPUT_FILE = "xmlInputFile";
    public static final String XML_INPUT_STRING = "xmlInputString";
    /**
     * Name of parameter that identifies a resource that can be retrieved as a stream from the
     * servlet context.
     */
    public static final String XML_INPUT_RESOURCE = "xmlInputResource";

    private Map<String, String> parameterMap = new HashMap<String, String>();
    private BaseVariantXMLSource source = new BaseVariantXMLSource();

    @Override
    public void configureSource(FilterConfig config) throws VariantSourceConfigurationException
    {
        ServletContext context = config.getServletContext();
        @SuppressWarnings("unchecked")
        Enumeration<String> parameterNames = config.getInitParameterNames();
        while (parameterNames.hasMoreElements())
        {
            String parameterName = parameterNames.nextElement();
            String parameterValue = config.getInitParameter(parameterName);
            parameterMap.put(parameterName, parameterValue);
        }

        if (parameterMap.containsKey(SAX_HANDLER))
        {
            try
            {
                Class<?> handlerClass = Class.forName(parameterMap.get(SAX_HANDLER));
                Object handlerObject = handlerClass.newInstance();
                source.setHandler((VariantSAXHandler) handlerObject);
            }
            catch (Exception ex)
            {
                throw new VariantSourceConfigurationException("Unable to construct SAX handler", ex);
            }
        }
        if (parameterMap.containsKey(XML_INPUT_RESOURCE))
        {
            source.setInputStream(context.getResourceAsStream(parameterMap.get(XML_INPUT_RESOURCE)));
        }
        else if (parameterMap.containsKey(XML_INPUT_FILE))
        {
            try
            {
                source.setInputStream(new FileInputStream(parameterMap.get(XML_INPUT_FILE)));
            }
            catch (FileNotFoundException exception)
            {
                throw new VariantSourceConfigurationException("Unable get XML file", exception);
            }
        }
        else if (parameterMap.containsKey(XML_INPUT_STRING))
        {
            source.setInputStream(new ByteArrayInputStream(parameterMap.get(XML_INPUT_STRING)
                    .getBytes()));
        }
        else
        {
            throw new VariantSourceConfigurationException("Input stream not set");
        }
    }

    @Override
    public List<Variant> getVariants() throws VariantSourceException
    {
        return source.getVariants();
    }
}
